/*
 *  Code for doing 3D transforms
 *  
 */

#include "3d.h"

void
set_v_tx(tx *tx, point *eye, point *coi)
/*{{{  */
{
   /*{{{  Pythag version from point*/
   
   double yh = (double)sqrt((double)(eye->x*eye->x + eye->z*eye->z));
   double cosy = eye->z/yh,
          siny = eye->x/yh;
   double xh = (double)sqrt((double)(eye->y*eye->y + yh*yh));
   double cosx = yh/xh,
          sinx = eye->y/xh;
   
   /*}}}  */

   /*{{{  Trig version using angles*/
   /*
   double sinx,siny,cosx,cosy;
   
   sincos(xa, &sinx, &cosx);
   sincos(ya, &siny, &cosy);
   */
   /*}}}  */
    
   /* 
    *  Create a transform to translate such that the point coi
    *  (the centre of interest) becomes the origin and rotate so 
    *  that the eye position is on the z axis.
    */
   
   tx->r11 = cosy;
   tx->r12 = 0.0;
   tx->r13 = -siny;
   
   tx->r21 = sinx*siny;
   tx->r22 = cosx;
   tx->r23 = sinx*cosy;
   
   tx->r31 = cosx*siny;
   tx->r32 = -sinx;
   tx->r33 = cosx*cosy;
   
   tx->dx  = -(coi->x*tx->r11 + coi->y*tx->r21 + coi->z*tx->r31);
   tx->dy  = -(coi->x*tx->r12 + coi->y*tx->r22 + coi->z*tx->r32);
   tx->dz  = -(coi->x*tx->r13 + coi->y*tx->r23 + coi->z*tx->r33);
}
/*}}}  */

void
set_ball_tx(tx *tx, double xa, double ya, point *p)
/*{{{  */
{
   /*{{{  Pythag version from point*/
   /*
   float yh = (float)sqrt((double)(p->x*p->x + p->z*p->z));
   float cosy = p->z/yh,
         siny = p->x/yh;
   float xh = (float)sqrt((double)(p->y*p->y + yh*yh));
   float cosx = yh/xh,
         sinx = p->y/xh;
   */
   /*}}}  */

   double sinx,siny,cosx,cosy;
   
   sincos(xa, &sinx, &cosx);
   sincos(ya, &siny, &cosy);
    
   /* 
    *  Create a transform to rotate by xa around x-axis,
    *  and by ya around y-axis, and translate to point p.
    */
   
   tx->r11 = cosy;
   tx->r12 = 0.0;
   tx->r13 = -siny;
   
   tx->r21 = sinx*siny;
   tx->r22 = cosx;
   tx->r23 = sinx*cosy;
   
   tx->r31 = cosx*siny;
   tx->r32 = -sinx;
   tx->r33 = cosx*cosy;
   
   tx->dx = p->x;
   tx->dy = p->y;
   tx->dz = p->z;
}
/*}}}  */

void
tx_point(point *p1, tx *tx, point *p2)
/*{{{  */
{
   float x,y,z; 
   
/*
   x = p1->x * tx->r11 + 
       p1->y * tx->r12 + 
       p1->z * tx->r13 + tx->dx;
                   
   y = p1->x * tx->r21 + 
       p1->y * tx->r22 + 
       p1->z * tx->r23 + tx->dy;
                   
   z = p1->x * tx->r31 + 
       p1->y * tx->r32 + 
       p1->z * tx->r33 + tx->dz;
*/       

   x = p1->x * tx->r11 + 
       p1->y * tx->r21 + 
       p1->z * tx->r31 + tx->dx;
                   
   y = p1->x * tx->r12 + 
       p1->y * tx->r22 + 
       p1->z * tx->r32 + tx->dy;
                   
   z = p1->x * tx->r13 + 
       p1->y * tx->r23 + 
       p1->z * tx->r33 + tx->dz;

   p2->x = x;   /* Copy in case p1 == p2 */
   p2->y = y;
   p2->z = z;
}
/*}}}  */

point*
add_point(point *p1, point *p2, point *res)
/*{{{  */
{
   res->x = p1->x + p2->x;
   res->y = p1->y + p2->y;
   res->z = p1->z + p2->z;

   return res;
}
/*}}}  */

point*
sub_point(point *p1, point *p2, point *res)
/*{{{  */
{
   res->x = p1->x - p2->x;
   res->y = p1->y - p2->y;
   res->z = p1->z - p2->z;

   return res;
}
/*}}}  */

void
identity_tx(tx *t)
/*{{{  */
{
   t->r11 = t->r22 = t->r33 = 1.0;
   t->r12 = t->r13 = t->r21 = t->r23 = t->r31 = t->r32 = 0.0;
   t->dx = t->dy = t->dz = 0.0;
}
/*}}}  */

void
multiply_tx(tx *t1, tx *t2, tx *res)
/*{{{  */
{
   /*
    *   Pre-multiply t1 by t2 giving res
    */
   
   res->r11 = t2->r11*t1->r11 + t2->r21*t1->r12 + t2->r31*t1->r13;
   res->r21 = t2->r11*t1->r21 + t2->r21*t1->r22 + t2->r31*t1->r23;
   res->r31 = t2->r11*t1->r31 + t2->r21*t1->r32 + t2->r31*t1->r33;
   res->dx  = t2->r11*t1->dx  + t2->r21*t1->dy  + t2->r31*t1->dz + t2->dx;
   
   res->r12 = t2->r12*t1->r11 + t2->r22*t1->r12 + t2->r32*t1->r13;
   res->r22 = t2->r12*t1->r21 + t2->r22*t1->r22 + t2->r32*t1->r23;
   res->r32 = t2->r12*t1->r31 + t2->r22*t1->r32 + t2->r32*t1->r33;
   res->dy  = t2->r12*t1->dx  + t2->r22*t1->dy  + t2->r32*t1->dz + t2->dy;
      
   res->r13 = t2->r13*t1->r11 + t2->r23*t1->r12 + t2->r33*t1->r13;
   res->r23 = t2->r13*t1->r21 + t2->r23*t1->r22 + t2->r33*t1->r23;
   res->r33 = t2->r13*t1->r31 + t2->r23*t1->r32 + t2->r33*t1->r33;
   res->dz  = t2->r13*t1->dx  + t2->r23*t1->dy  + t2->r33*t1->dz + t2->dz;
}
/*}}}  */

void
rotate_tx(tx *t, char axis, double angle)
/*{{{  */
{
   /*
    * Rotate the transform about the given axis
    * by the specified angle (radians)
    */
    
   tx     rot,res;
   double s,c;
   
   sincos(angle,&s,&c);
   
   identity_tx(&rot);
   
   switch(axis)
   {
   case 'x':
      rot.r22 = rot.r33 = c;
      rot.r23 = s;
      rot.r32 = -s;
      break;
      
   case 'y':
      rot.r11 = rot.r33 = c;
      rot.r31 = s;
      rot.r13 = -s;
      break;
      
   case 'z':
      rot.r11 = rot.r22 = c;
      rot.r12 = s;
      rot.r21 = -s;
      break;
   }
   
   multiply_tx(t, &rot, &res);
   
   *t = res;
}
/*}}}  */

void
rotate_tx_sc(tx *t, char axis, double s, double c)
/*{{{  */
{
   /*
    * Rotate the transform about the given axis
    * by the angle whose sin and cosine are specified
    */
    
   tx     rot,res;
   
   identity_tx(&rot);
   
   switch(axis)
   {
   case 'x':
      rot.r22 = rot.r33 = c;
      rot.r23 = s;
      rot.r32 = -s;
      break;
      
   case 'y':
      rot.r11 = rot.r33 = c;
      rot.r31 = s;
      rot.r13 = -s;
      break;
      
   case 'z':
      rot.r11 = rot.r22 = c;
      rot.r12 = s;
      rot.r21 = -s;
      break;
   }
   
   multiply_tx(t, &rot, &res);
   
   *t = res;
}
/*}}}  */

void
pre_offset_tx(tx *t, point *p)
/*{{{  */
{
   /*
    * Replace any existing pre-offset component
    * of the transform with an offset defined by p
    */
   
   t->dx = -(t->r11*p->x + t->r21*p->y + t->r31*p->z);
   t->dy = -(t->r12*p->x + t->r22*p->y + t->r32*p->z);
   t->dz = -(t->r13*p->x + t->r23*p->y + t->r33*p->z);
}
/*}}}  */

void
offset_tx(tx *t, point *p)
/*{{{  */
{
   /*
    * Add a relative translation defined by p
    */
   
   t->dx += p->x;
   t->dy += p->y;
   t->dz += p->z;
}
/*}}}  */

